import logging
import time
from functools import lru_cache
from typing import Dict, List, Sequence, Set, Tuple, cast

from fastapi import Depends, HTTPException, status
from ldap3.utils.conv import escape_filter_chars

from id_broker_common.kelvin import get_kelvin_obj, kelvin_session
from id_broker_common.ldap_access import get_users
from id_broker_common.models import RawUser
from id_broker_common.pseudonyms_udm_rest import (
    IdBrokerPseudonymisationError,
    get_service_provider_attribute,
)
from id_broker_common.utils import ldap_credentials, ldap_settings, remove_prefix
from ucsschool.apis.models import OAuth2Token
from ucsschool.apis.utils import LDAPAccess, auth_manager, get_logger
from ucsschool.kelvin.client import SchoolClass, SchoolClassResource, Session, User, UserResource

from .models import ClassMembers, NonEmptyStr, NoStarStr, Student, Teacher

logger: logging.Logger = get_logger()


# Extracted to variable to allow dep overrides in FastAPI, which cannot resolve anonymous dependencies
# TODO: import PLUGIN_NAME without circular import error:
token_dep = auth_manager("self_disclosure", OAuth2Token)


# # TODO: these caches should have a TTL, so changes to the LDAP data get recognized
# # {user_name: {ou: set(role, ...)}}
# user_name_to_role_per_school_cache: Dict[str, Dict[str, Iterable[str]]] = {}
# # {user_name: {service_provider: pseudonym}}
# user_name_to_pseudonyms_cache: Dict[str, Dict[str, str]] = {}


def service_provider_from_token(token: OAuth2Token = Depends(token_dep)) -> str:
    return token.client_id


def connecting_user_pseudonym_from_token(token: OAuth2Token = Depends(token_dep)) -> str:
    return token.sub


@lru_cache(maxsize=1024)
def role_split(ucsschool_role: str) -> Tuple[str, str, str]:
    """Cached: same ucsschool_role string results in same tuple."""
    role, context_type, context_name = ucsschool_role.split(":")
    return role, context_type, context_name


@lru_cache(maxsize=1024)
def roles_in_ucsschool_roles(ucsschool_roles: Tuple[str, ...], school_name: str) -> Set[str]:
    """Cached: same list of ucsschool_roles strings (hashable->sorted tuple) returns same set of roles."""
    return {
        role
        for role, context_type, context_name in [
            role_split(ucsschool_role) for ucsschool_role in ucsschool_roles
        ]
        if context_type == "school" and school_name == context_name
    }


def roles_in_school(ucsschool_roles: Sequence[str], school_name: str) -> Set[str]:
    """Helper to get only the roles that are defined for a user within a specific school"""
    try:
        return roles_in_ucsschool_roles(tuple(sorted(ucsschool_roles)), school_name)
    except ValueError:
        logger.error("Bad value in 'ucsschool_roles': %r", ucsschool_roles)
        return set()


async def extract_members(
    schoolclass: SchoolClass,
    service_provider: str,
    *,
    list_students=True,
    list_teachers=True,
) -> ClassMembers:
    """
    Helper to get users of a school_class split into teachers and students

    :raises id_broker_common.udm_rest.IdBrokerPseudonymisationError: if `service_provider` could not be
        loaded from / found in `settings/data` UDM object
    """
    teachers = []
    students = []
    sp_attribute = await get_service_provider_attribute(service_provider)
    t0 = time.time()
    users: List[RawUser] = await get_users(schoolclass.users)
    logger.debug("Retrieved %d users in %0.2f sec from LDAP.", len(users), time.time() - t0)
    for user in users:
        school_authority_prefix = f"{user.source_uid}-"
        if not user.synonyms.get(sp_attribute):
            logger.warning(
                "Ignoring user %r without pseudonym in %r (for service provider %r).",
                user.username,
                sp_attribute,
                service_provider,
            )
            continue
        roles = roles_in_school(user.ucsschool_roles, schoolclass.school)
        if list_students and "student" in roles:
            student = Student(
                user_id=user.synonyms[sp_attribute],
                username=remove_prefix(user.username, school_authority_prefix),
                firstname=user.firstname,
                lastname=user.lastname,
            )
            students.append(student)
        if list_teachers and "teacher" in roles:
            teacher = Teacher(
                user_id=user.synonyms[sp_attribute],
                username=remove_prefix(user.username, school_authority_prefix),
            )
            teachers.append(teacher)
    return ClassMembers(students=students, teachers=teachers)


async def get_user(pseudonym: str, service_provider: str, session: Session) -> User:
    """
    Retrieve user.

    :param str pseudonym: a users pseudonym
    :param str service_provider: the service provider for which to search user with the given pseudonym
    :param Session session: Kelvin client session instance
    :return: Kelvin User object
    :raises HTTPException: 404 if not found
    :raises id_broker_common.udm_rest.IdBrokerPseudonymisationError: if `service_provider` could not be
        loaded from / found in `settings/data` UDM object
    """
    sp_attribute = await get_service_provider_attribute(service_provider)
    search_params = {sp_attribute: pseudonym}
    async for user in UserResource(session=session).search(**search_params):
        # We expect entryUUID values (the only individual element in a pseudonym) to be unique not only
        # for a domain, but globally. If that'd prove to be wrong, the search parameters should either
        # include the school authority (stored in 'ucsschoolSourceUID') or collect all results of the
        # search and raise an error if more than 1 is found.
        return user
    else:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"No user was found with pseudonym {pseudonym!r} for service provider "
            f"{service_provider!r}.",
        )


async def get_connected_user(
    service_provider: str = Depends(service_provider_from_token),
    pseudonym_of_connecting_user: str = Depends(connecting_user_pseudonym_from_token),
    session: Session = Depends(kelvin_session),
) -> User:
    logger.debug(
        "service_provider=%r pseudonym_of_connecting_user=%r",
        service_provider,
        pseudonym_of_connecting_user,
    )
    try:
        return await get_user(pseudonym_of_connecting_user, service_provider, session)
    except (HTTPException, IdBrokerPseudonymisationError) as exc:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED) from exc


async def get_school_class(pseudonym: str, service_provider: str, session: Session) -> SchoolClass:
    """
    Retrieve school class.

    :param str pseudonym: a school class' pseudonym
    :param str service_provider: the service provider for which to search group with the given pseudonym
    :param Session session: Kelvin client session instance
    :return: Kelvin SchoolClass object
    :raises HTTPException: 404 if not found
    :raises id_broker_common.udm_rest.IdBrokerPseudonymisationError: if `service_provider` could not be
        loaded from / found in `settings/data` UDM object
    """
    sp_attribute = await get_service_provider_attribute(service_provider)
    attrs = await _get_obj_attr_by_pseudonym(pseudonym, sp_attribute, "ucsschoolGroup", "cn")
    school = attrs["school"]
    name = remove_prefix(attrs["name"], f"{school}-")
    return cast(
        SchoolClass,
        await get_kelvin_obj(
            SchoolClassResource(session=session),
            name=name,
            school=school,
        ),
    )


async def _get_obj_attr_by_pseudonym(
    pseudonym: NoStarStr,
    sp_attribute: str,
    object_class: NonEmptyStr,
    name_attr: NonEmptyStr,
) -> Dict[str, str]:
    """Get 'name' and 'school' from `name_attr` and `ucsschoolRole` of LDAP object."""
    ldap = LDAPAccess(ldap_settings(), ldap_credentials())
    filter_s = f"(&(objectClass={object_class})" f"({sp_attribute}={escape_filter_chars(pseudonym)}))"
    results = await ldap.search(filter_s, attributes=[name_attr, "ucsschoolRole"])
    if len(results) == 1:
        result = results[0]
        name = result[name_attr].value
        if isinstance(result.ucsschoolRole.value, str):
            school_value = result.ucsschoolRole.value
        else:
            school_value = [s for s in result.ucsschoolRole.value if ":school:" in s][0]
        school = school_value.split(":")[-1]
        return {"name": name, "school": school}
    elif len(results) > 1:
        logger.error("More than 1 result when searching LDAP with filter %r: %r.", filter_s, results)
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"Error retrieving object with pseudonym {pseudonym!r}.",
        )
    else:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Object with pseudonym {pseudonym!r} not found.",
        )
